#pragma once

#include "openpifpaf/utils.hpp"
#include "openpifpaf/decoder/cifcaf.hpp"
#include "openpifpaf/decoder/cifdet.hpp"

#include "openpifpaf/decoder/utils/caf_scored.hpp"
#include "openpifpaf/decoder/utils/cif_hr.hpp"
#include "openpifpaf/decoder/utils/cif_seeds.hpp"
#include "openpifpaf/decoder/utils/nms_keypoints.hpp"
#include "openpifpaf/decoder/utils/occupancy.hpp"
