import torch

from . import nms

# pylint: disable=invalid-name
CafScored = torch.classes.openpifpaf_decoder_utils.CafScored
CifHr = torch.classes.openpifpaf_decoder_utils.CifHr
CifSeeds = torch.classes.openpifpaf_decoder_utils.CifSeeds
CifDetSeeds = torch.classes.openpifpaf_decoder_utils.CifDetSeeds
Occupancy = torch.classes.openpifpaf_decoder_utils.Occupancy
